'''
Created on Feb 14, 2020

@author: dwolz
'''

import time
import math
from scipy.optimize import minimize, differential_evolution, dual_annealing, shgo
from fcmaes.optimizer import dtime, random_x, typical, scale

from fcmaes import astro

def test_scipy_minimize(problem, num):
    best = math.inf
    t0 = time.perf_counter();
    for i in range(num):
        guess = random_x(problem.bounds.lb, problem.bounds.ub)
        ret = minimize(problem.fun, x0 = guess, bounds = problem.bounds)
        best = min(ret.fun, best)
        print("{0}: time = {1:.1f} best = {2:.1f} f(xmin) = {3:.1f}"
              .format(i+1, dtime(t0), best, ret.fun))

def test_shgo(problem, num):
    best = math.inf
    lb = problem.bounds.lb
    ub = problem.bounds.ub
    t0 = time.perf_counter();
    for i in range(num):
        ret = shgo(problem.fun, bounds = list(zip(lb, ub)), 
                   n = 300, sampling_method = 'sobol')
        best = min(ret.fun, best)
        print("{0}: time = {1:.1f} best = {2:.1f} f(xmin) = {3:.1f}"
              .format(i+1, dtime(t0), best, ret.fun))

def test_dual_annealing(problem, num):
    best = math.inf
    lb = problem.bounds.lb
    ub = problem.bounds.ub
    t0 = time.perf_counter();
    for i in range(num):
        ret = dual_annealing(problem.fun, bounds = list(zip(lb, ub)))
        best = min(ret.fun, best)
        print("{0}: time = {1:.1f} best = {2:.1f} f(xmin) = {3:.1f}"
              .format(i+1, dtime(t0), best, ret.fun))

def test_differential_evolution(problem, num):
    best = math.inf
    t0 = time.perf_counter();
    for i in range(num):
        ret = differential_evolution(problem.fun, bounds = problem.bounds)
        best = min(ret.fun, best)
        print("{0}: time = {1:.1f} best = {2:.1f} f(xmin) = {3:.1f}"
              .format(i+1, dtime(t0), best, ret.fun))

def test_cma_original(problem, num):
    import cma
    best = math.inf
    lb = problem.bounds.lb
    ub = problem.bounds.ub
    t0 = time.perf_counter();
    for i in range(num):
        guess = random_x(problem.bounds.lb, problem.bounds.ub)
        es = cma.CMAEvolutionStrategy(guess, 1.0,  
                                      {'bounds': [lb, ub], 'popsize': 32, 
                                        'typical_x': typical(lb, ub),
                                        'scaling_of_variables': scale(lb, ub),
                                        'verbose': -1, 'verb_disp': -1})
        for j in range(100000):
            X, Y = es.ask_and_eval(problem.fun)
            es.tell(X, Y)
            if es.stop():
                break 
        best = min(es.result.fbest, best)
        print("{0}: time = {1:.1f} best = {2:.1f} f(xmin) = {3:.1f}"
              .format(i+1, dtime(t0), best, es.result.fbest))

from fcmaes import cmaes

def test_cma_python(problem, num):
    best = math.inf
    t0 = time.perf_counter();
    for i in range(num):
        ret = cmaes.minimize(problem.fun, bounds = problem.bounds)
        best = min(ret.fun, best)
        print("{0}: time = {1:.1f} best = {2:.1f} f(xmin) = {3:.1f}"
              .format(i+1, dtime(t0), best, ret.fun))
        
from fcmaes import cmaescpp

def test_cma_cpp(problem, num):
    best = math.inf
    t0 = time.perf_counter();
    for i in range(num):
        ret = cmaescpp.minimize(problem.fun, bounds = problem.bounds)
        best = min(ret.fun, best)
        print("{0}: time = {1:.1f} best = {2:.1f} f(xmin) = {3:.1f}"
              .format(i+1, dtime(t0), best, ret.fun))

from fcmaes import retry

def test_retry_python(problem, num):
    ret = retry.minimize(problem.fun, bounds = problem.bounds, num_retries = num, 
                   max_evaluations = 50000)

def test_retry_cpp(problem, num):
    ret = retry.minimize(problem.fun, bounds = problem.bounds, num_retries = num, 
                   max_evaluations = 50000, useCpp = True)

from fcmaes import advretry

def test_advretry_python(problem, value_limit, num):
    best = math.inf
    t0 = time.perf_counter();    
    for i in range(num):
        ret = advretry.minimize(problem.fun, bounds = problem.bounds, num_retries = 4000, 
                   value_limit = value_limit, logger = None)
        best = min(ret.fun, best)
        print("{0}: time = {1:.1f} best = {2:.1f} f(xmin) = {3:.1f}"
              .format(i+1, dtime(t0), best, ret.fun))

def test_advretry_cpp(problem, value_limit, num):
    best = math.inf
    t0 = time.perf_counter();    
    for i in range(num):
        ret = advretry.minimize(problem.fun, bounds = problem.bounds, num_retries = 4000, 
                   value_limit = value_limit, logger = None, useCpp = True)
        best = min(ret.fun, best)
        print("{0}: time = {1:.1f} best = {2:.1f} f(xmin) = {3:.1f}"
              .format(i+1, dtime(t0), best, ret.fun))

if __name__ == '__main__':
    problem = astro.Gtoc1()
#     problem = astro.Cassini1()
#     problem = astro.Messenger()
#     problem = astro.MessFull()
 
#     test_scipy_minimize(problem, 1000)
#     test_shgo(problem, 2)
#     test_dual_annealing(problem, 100)
#     test_differential_evolution(problem, 100)
#     test_cma_original(problem, 100)
    test_cma_python(problem, 100)
#     test_cma_cpp(problem, 100)
#     test_retry_python(problem, 5000)
#     test_retry_cpp(problem, 5000)
#     test_advretry_python(problem, -1000000, 10)
#     test_advretry_cpp(problem, -1000000, 10)




'''
Created on Feb 9, 2020

@author: Dietmar Wolz
'''

from fcmaes.astro import MessFull, Messenger, Gtoc1, Cassini1
from fcmaes.optimizer import  Optimizer, logger
from fcmaes.advretry import Store, retry, minimize

def messengerFullLoop():    
    while True:    
        problem = MessFull()
        logger.info(problem.name + ' cmaes c++')
        ret = minimize(problem.fun, bounds=problem.bounds, num_retries = 40000, 
                       max_evaluations = 50000, value_limit = 10.0, useCpp = True)

def test_all(num_retries = 4000, num = 20):
     
    problem = Cassini1()
    logger.info(problem.name + ' cmaes c++')
    for i in range(num):
        ret = minimize(problem.fun, bounds=problem.bounds, num_retries = num_retries, 
                       useCpp = True, value_limit = 12.0)
 
    problem = Messenger()
    logger.info(problem.name + ' cmaes c++')
    for i in range(num):
        ret = minimize(problem.fun, bounds=problem.bounds, num_retries = num_retries, 
                       useCpp = True, value_limit = 10.0)
 
    problem = Gtoc1()
    logger.info(problem.name + ' cmaes c++')
    for i in range(num):
        ret = minimize(problem.fun, bounds=problem.bounds, num_retries = num_retries, 
                       useCpp = True, value_limit = -1000000)
 
    problem = MessFull()
    logger.info(problem.name + ' cmaes c++')
    for i in range(num):
        ret = minimize(problem.fun, bounds=problem.bounds, num_retries = num_retries, 
                       useCpp = True, value_limit = 10.0)


    problem = Cassini1()
    logger.info(problem.name + ' cmaes python')
    for i in range(num):
        ret = minimize(problem.fun, bounds=problem.bounds, num_retries = num_retries, 
                       useCpp = False, value_limit = 12.0)
 
    problem = Messenger()
    logger.info(problem.name + ' cmaes python')
    for i in range(num):
        ret = minimize(problem.fun, bounds=problem.bounds, num_retries = num_retries, 
                       useCpp = False, value_limit = 10.0)

    problem = Gtoc1()
    logger.info(problem.name + ' cmaes python')
    for i in range(num):
        ret = minimize(problem.fun, bounds=problem.bounds, num_retries = num_retries, 
                       useCpp = False, value_limit = -1000000)

    problem = MessFull()
    logger.info(problem.name + ' cmaes python')
    for i in range(num):
        ret = minimize(problem.fun, bounds=problem.bounds, num_retries = num_retries, 
                       useCpp = False, value_limit = 10.0)
        

    problem = Cassini1()
    logger.info(problem.name + ' dual annealing')
    for i in range(num):
        store = Store(problem.bounds)
        optimizer  = Optimizer(store, 0)
        ret = retry(problem.fun, store, optimizer.dual_annealing, num_retries, 20.0)

    problem = Messenger()
    logger.info(problem.name + ' dual annealing')
    for i in range(num):
        store = Store(problem.bounds)
        optimizer  = Optimizer(store, 0)
        ret = retry(problem.fun, store, optimizer.dual_annealing, num_retries, 20.0)

    problem = Gtoc1()
    logger.info(problem.name + ' dual annealing')
    for i in range(num):
        store = Store(problem.bounds)
        optimizer  = Optimizer(store, 0)
        ret = retry(problem.fun, store, optimizer.dual_annealing, num_retries, -200000)

    problem = MessFull()
    logger.info(problem.name + ' dual annealing')
    for i in range(num):
        store = Store(problem.bounds)
        optimizer  = Optimizer(store, 0)
        ret = retry(problem.fun, store, optimizer.dual_annealing, num_retries, 20.0)
   
        
    problem = Cassini1()
    logger.info(problem.name + ' differential evolution')
    for i in range(num):
        store = Store(problem.bounds)
        optimizer  = Optimizer(store, 0)
        ret = retry(problem.fun, store, optimizer.differential_evolution, num_retries, 20.0)

    problem = Messenger()
    logger.info(problem.name + ' differential evolution')
    for i in range(num):
        store = Store(problem.bounds)
        optimizer  = Optimizer(store, 0)
        ret = retry(problem.fun, store, optimizer.differential_evolution, num_retries, 20.0)

    problem = Gtoc1()
    logger.info(problem.name + ' differential evolution')
    for i in range(num):
        store = Store(problem.bounds)
        optimizer  = Optimizer(store, 0)
        ret = retry(problem.fun, store, optimizer.differential_evolution, num_retries, -200000)

    problem = MessFull()
    logger.info(problem.name + ' differential evolution')
    for i in range(num):
        store = Store(problem.bounds)
        optimizer  = Optimizer(store, 0)
        ret = retry(problem.fun, store, optimizer.differential_evolution, num_retries, 20.0)


def main():
    test_all()
    #messengerFullLoop()

if __name__ == '__main__':
    main()


'''
Created on Jan 10, 2020

@author: Dietmar Wolz
'''

import random
import time
import math
import os
import ctypes as ct
import numpy as np
from numpy.random import Generator, MT19937, SeedSequence
from scipy.optimize._constraints import new_bounds_to_old
from scipy.optimize import OptimizeResult, Bounds
import multiprocessing as mp
from multiprocessing import Process

from fcmaes.optimizer import Optimizer, dtime, logger, seed_random

os.environ['MKL_DEBUG_CPU_TYPE'] = '5'
os.environ['MKL_NUM_THREADS'] = '1'
os.environ['OPENBLAS_NUM_THREADS'] = '1'

def minimize(fun, 
             bounds = None, 
             value_limit = math.inf,
             num_retries = 1000,
             logger = logger,
             workers = mp.cpu_count(),
             popsize = 31, 
             max_evaluations = 50000, 
             useCpp = False,
             stop_fittness = None,
             ):   
    """Minimization of a scalar function of one or more variables using parallel 
     CMA-ES retry.
     
    Parameters
    ----------
    fun : callable
        The objective function to be minimized.
            ``fun(x, *args) -> float``
        where ``x`` is an 1-D array with shape (n,) and ``args``
        is a tuple of the fixed parameters needed to completely
        specify the function.
    bounds : sequence or `Bounds`, optional
        Bounds on variables for L-BFGS-B, TNC, SLSQP and
        trust-constr methods. There are two ways to specify the bounds:
            1. Instance of the `scipy.Bounds` class.
            2. Sequence of ``(min, max)`` pairs for each element in `x`. None
               is used to specify no bound.
    value_limit : float, optional
        Upper limit for CMA-ES optimized function values to be stored. 
    num_retries : int, optional
        Number of CMA-ES retries.    
    logger : logger, optional
        logger for log output of the retry mechanism. If None, logging
        is switched off. Default is a logger which logs both to stdout and
        appends to a file ``optimizer.log``.
    workers : int, optional
        number of parallel processes used. Default is mp.cpu_count()
    popsize = int, optional
        CMA-ES population size used for all CMA-ES runs.
    max_evaluations : int, optional
        Forced termination of all CMA-ES runs after ``max_evaluations`` 
        function evaluations.
    useCpp : bool, optional
        Flag indicating use of the C++ CMA-ES implementation. Default is `False` - 
        use of the Python CMA-ES implementation
    stop_fittness : float, optional 
         Limit for fitness value. CMA-ES runs terminate if this value is reached. 
    
    Returns
    -------
    res : scipy.OptimizeResult
        The optimization result is represented as an ``OptimizeResult`` object.
        Important attributes are: ``x`` the solution array, 
        ``fun`` the best function value, ``nfev`` the number of function evaluations,
        ``nit`` the number of CMA-ES iterations, ``status`` the stopping critera and
        ``success`` a Boolean flag indicating if the optimizer exited successfully. """
    
    store = Store(bounds, max_evaluations = max_evaluations, logger = logger)
    optimizer  = Optimizer(store, popsize, stop_fittness)
    optimize = optimizer.cma_cpp if useCpp else optimizer.cma_python
    return retry(fun, store, optimize, num_retries, value_limit, workers)
                
def retry(fun, store, optimize, num_retries, value_limit = math.inf, workers=mp.cpu_count()):
    sg = SeedSequence()
    rgs = [Generator(MT19937(s)) for s in sg.spawn(workers)]
    proc=[Process(target=retry_loop,
            args=(pid, rgs, fun, store, optimize, num_retries, value_limit)) for pid in range(workers)]
    [p.start() for p in proc]
    [p.join() for p in proc]
    store.sort()
    store.dump()
    return OptimizeResult(x=store.get_x(0), fun=store.get_y(0), nfev=store.get_count_evals(), success=True)
 
def retry_loop(pid, rgs, fun, store, optimize, num_retries, value_limit):
    seed_random() # make sure cpp random generator for this process is initialized properly
    lower = store.lower
    while store.get_count_runs_incr() < num_retries:               
        sol, y, evals = optimize(fun, None, Bounds(store.lower, store.upper), 
                                 [random.uniform(0.05, 0.1)]*len(lower), rgs[pid])
        store.add_result(y, sol, evals, value_limit)
        if pid == 0:
            store.dump()

def convertBounds(bounds):
    if bounds is None:
        raise ValueError('bounds need to be defined')
    if isinstance(bounds, Bounds):
        limits = np.array(new_bounds_to_old(bounds.lb,
                                                 bounds.ub,
                                                 len(bounds.lb)),
                               dtype=float).T
    else:
        limits = np.array(bounds, dtype='float').T
    if (np.size(limits, 0) != 2 or not
            np.all(np.isfinite(limits))):
        raise ValueError('bounds should be a sequence containing '
                         'real valued (min, max) pairs for each value'
                         ' in x')
    return limits[0], limits[1]


class Store(object):
    """thread safe storage for optimization retry results."""
       
    def __init__(self, 
                 bounds, # bounds of the objective function arguments
                 max_evaluations = 50000, # maximum evaluation count
                 check_interval = 10, # sort evaluation memory after check_interval iterations
                 capacity = 500, # capacity of the evaluation store
                 logger = logger # if None logging is switched off
                ):    
        self.lower, self.upper = convertBounds(bounds)
        self.logger = logger
        self.max_evals = max_evaluations   
        self.capacity = capacity
        self.check_interval = check_interval
        self.dim = len(self.lower)
        self.delta = []
        for k in range(self.dim):
            self.delta.append(self.upper[k] - self.lower[k])
        
        #shared between processes
        self.add_mutex = mp.Lock()    
        self.xs = mp.RawArray(ct.c_double, self.capacity * self.dim)
        self.ys = mp.RawArray(ct.c_double, self.capacity)  
        self.count_evals = mp.RawValue(ct.c_long, 0)   
        self.count_runs = mp.RawValue(ct.c_int, 0) 
        self.num_stored = mp.RawValue(ct.c_int, 0) 
        self.num_sorted = mp.RawValue(ct.c_int, 0)  
        self.count_stat_runs = mp.RawValue(ct.c_int, 0)  
        self.t0 = time.perf_counter()
        self.mean = mp.RawValue(ct.c_double, 0) 
        self.qmean = mp.RawValue(ct.c_double, 0) 
        self.best_y = mp.RawValue(ct.c_double, math.inf) 
    
    def eval_num(self):
        return self.max_evals
                                             
    def replace(self, i, y, xs):
        self.set_y(i, y)
        self.set_x(i, xs)
             
    def sort(self): # sort all entries to make room for new ones, determine best and worst
        """sorts all store entries, keep only the 90% best to make room for new ones."""
        ns = self.num_stored.value
        ys = np.asarray(self.ys[:ns])
        yi = ys.argsort()
        sortRuns = []
        for i in range(len(yi)):
            y = ys[yi[i]]
            xs = self.get_x(yi[i])
            sortRuns.append((y, xs))
        numStored = min(len(sortRuns),int(0.9*self.capacity)) # keep 90% best 
        for i in range(numStored):
            self.replace(i, sortRuns[i][0], sortRuns[i][1])
        self.num_sorted.value = numStored  
        self.num_stored.value = numStored     
        return numStored        
            
    def add_result(self, y, xs, evals, limit=math.inf):
        """registers an optimization result at the score."""
        with self.add_mutex:
            self.incr_count_evals(evals)
            if y < limit:  
                self.count_stat_runs.value += 1
                if y < self.best_y.value:
                    self.dump()
                    self.best_y.value = y
                if self.num_stored.value >= self.capacity-1:
                    self.sort()
                cnt = self.count_stat_runs.value
                diff = y - self.mean.value
                self.qmean.value += (cnt - 1) * diff*diff / cnt;
                self.mean.value += diff / cnt
                ns = self.num_stored.value
                self.num_stored.value = ns + 1
                self.replace(ns, y, xs)
            
    def get_x(self, pid):
        return self.xs[pid*self.dim:(pid+1)*self.dim]
    
    def get_y(self, pid):
        return self.ys[pid]

    def get_ys(self):
        return self.ys[:self.num_stored.value]
             
    def get_y_mean(self):
        return self.mean.value
    
    def get_y_standard_dev(self):
        cnt = self.get_count_runs()
        return 0 if cnt <= 0 else math.sqrt(self.qmean.value / cnt)

    def get_count_evals(self):
        return self.count_evals.value
 
    def get_count_runs(self):
        return self.count_runs.value

    def get_count_runs_incr(self):
        with self.add_mutex:
            runs = self.count_runs.value
            self.count_runs.value += 1
            return runs

    def set_x(self, pid, xs):
        self.xs[pid*self.dim:(pid+1)*self.dim] = xs[:]
       
    def set_y(self, pid, y):
        self.ys[pid] = y    
        
    def incr_count_evals(self, evals):
        if self.count_runs.value % self.check_interval == self.check_interval-1:
            self.sort()
        self.count_evals.value += evals
            
    def dump(self):
        """logs the current status of the store if logger defined."""
        if self.logger is None:
            return
        Ys = self.get_ys()
        vals = []
        for i in range(min(20, len(Ys))):
            vals.append(round(Ys[i],2))     
        dt = dtime(self.t0)   
                 
        message = '{0} {1} {2} {3} {4:.4f} {5:.2f} {6:.2f} {7!s} {8!s}'.format(
            dt, int(self.count_evals.value / dt), self.count_runs.value, self.count_evals.value, \
                self.best_y.value, self.get_y_mean(), self.get_y_standard_dev(), vals, self.get_x(0))
        self.logger.info(message)



'''
Created on Feb 9, 2020

@author: Dietmar Wolz
'''

import numpy as np
from scipy.optimize import minimize, shgo, differential_evolution, dual_annealing, Bounds
import sys
import time
import logging
import random

from fcmaes import cmaes
from fcmaes import cmaescpp 

#default logger used by the parallel retry. Logs both to stdout and into a file
logging.basicConfig(filename='optimizer.log', level=logging.INFO, format='%(message)s')
stdout_handler = logging.StreamHandler(sys.stdout)
stdout_handler.setLevel(logging.INFO)
logger = logging.getLogger('optimizer')
logger.addHandler(stdout_handler)

def eprint(*args, **kwargs):
    """print message to stderr."""
    print(*args, file=sys.stderr, **kwargs)

def scale(lower, upper):
    """scaling = 0.5 * difference of the bounds."""
    return 0.5 * (np.asarray(upper) - np.asarray(lower))

def typical(lower, upper):
    """typical value = mean of the bounds."""
    return 0.5 * (np.asarray(upper) + np.asarray(lower))

def fitting(guess, lower, upper):
    """fit a guess into the bounds."""
    return np.minimum(np.asarray(upper), np.maximum(np.asarray(guess), np.asarray(lower)))

def is_terminate(runid, iterations, val):
    """dummy is_terminate call back."""
    return False    

def random_x(lower, upper):
    """feasible random value uniformly distributed inside the bounds."""
    lower = np.asarray(lower)
    upper = np.asarray(upper)
    return lower + np.multiply(upper - lower, np.random.rand(lower.size))
    
def dtime(t0):
    """time since t0."""
    return round(time.perf_counter() - t0, 2)

def seed_random():    
    """makes sure the c++ random generator for this process is initialized properly"""
    if sys.platform.startswith('linux'):
        cmaescpp.seed_random() 

class Optimizer(object):
    """Provides different optimization methods for use with parallel retry."""
       
    def __init__(self, store, popsize = 31, stop_fittness = None):        
        self.popsize = popsize
        self.stop_fittness = stop_fittness
        # store provides the (changing) upper limit of the number of function evaluations
        self.store = store 
         
    def cma_python(self, fun, guess, bounds, sdevs, rg): 
        """CMA_ES Python implementation."""
        ret = cmaes.minimize(fun, bounds, guess,
                input_sigma=sdevs, max_evaluations=self.store.eval_num(), 
                popsize=self.popsize, stop_fittness = self.stop_fittness,
                rg=rg, runid=self.store.get_count_runs())     
        return ret.x, ret.fun, ret.nfev

    def cma_cpp(self, fun, guess, bounds, sdevs, rg):
        """CMA_ES C++ implementation."""
        ret = cmaescpp.minimize(fun, bounds, guess,
                input_sigma=sdevs, max_evaluations=self.store.eval_num(), 
                popsize=self.popsize, stop_fittness = self.stop_fittness,
                rg=rg, runid=self.store.get_count_runs())     
        return ret.x, ret.fun, ret.nfev
    
    def dual_annealing(self, fun, guess, bounds, sdevs, rg):
        """scipy dual_annealing."""
        ret = dual_annealing(fun, bounds=list(zip(bounds.lb, bounds.ub)), 
                             maxfun=self.store.eval_num(), 
                             seed=random.randint(0, 2**32 - 1))
        return ret.x, ret.fun, ret.nfev

    def differential_evolution(self, fun, guess, bounds, sdevs, rg):
        """scipy differential_evolution."""
        popsize = 15 # default value for differential_evolution
        maxiter = int(self.store.eval_num() / (popsize * len(bounds.lb)) - 1)
        ret = differential_evolution(fun, bounds=bounds, maxiter=maxiter,
                                      seed=random.randint(0, 2**32 - 1))
        return ret.x, ret.fun, ret.nfev
    
    def minimize(self, fun, guess, bounds, sdevs, rg):
        """scipy minimize."""
        ret = minimize(fun, bounds=bounds)
        return ret.x, ret.fun, ret.nfev
 
    def shgo(self, fun, guess, bounds, sdevs, rg):
        """scipy shgo."""
        ret = shgo(fun, bounds=list(zip(bounds.lb, bounds.ub)), 
                   options={'maxfev': self.store.eval_num()})
        return ret.x, ret.fun, ret.nfev
